unit GutenbergTreeView;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, RxRichEd;

type
  Chapter = record
    TextIndex : integer;
    Number    : smallint;
  end;
  PChapter = ^Chapter;
  TGutenbergTreeView = class(TTreeView)
    procedure AddNodes(Memo : TRxRichEdit);
    procedure SaveToFile(const FileName: string);
    procedure LoadFromFile(const FileName: string);
  private
    FFileName : String;
    HasChapters : boolean;
    procedure SaveToStream(Stream: TStream);
    procedure LoadTreeFromStream(Stream: TStream);
    function GetBufStart(Buffer: PChar; var Level: Integer): PChar;
    procedure SaveTreeToStream(Stream: TStream);
    procedure LoadFromStream(Stream: TStream);
  public
    function GetIndex : integer;
    function GetChapter(CharNo : integer) : TTreeNode;
    function AddBookMark(Node : TTreeNode; CharNo : integer; Bookmark : String) : TTreeNode;
    procedure DeleteBookMark(Node : TTreeNode);
  published
    property FileName : String read FFileName write FFileName;
  end;

procedure Register;

implementation

procedure TGutenbergTreeView.AddNodes(Memo : TRxRichEdit);
var IndexFound, NewIndex: integer;
    First : TTreeNode;
    j     : smallint;
    LT    : integer;
    PointChapter :PChapter;
begin
  IndexFound := 0;
  LT := Length(Memo.Text);
  j := 1;
  First := Items.GetFirstNode;
  repeat
    NewIndex := IndexFound;
    repeat
      NewIndex := Memo.FindText('CHAPTER', NewIndex+1,
          LT-NewIndex-1, []);
      Memo.SelStart := NewIndex;
      Memo.SelLength := Length('CHAPTER');
    until (NewIndex = -1) or (Memo.SelText = 'CHAPTER') or (Memo.SelText = 'Chapter');
    if NewIndex <> -1 then begin
      IndexFound := NewIndex+1;
      New(PointChapter);
      PointChapter^.Number := j;
      PointChapter^.TextIndex := NewIndex;
      Selected := Items.AddObject(First, 'CHAPTER '+IntToStr(j), PointChapter);
      inc(j);
    end;
  until NewIndex = -1;
  if (Items.Count > 0) then
     hasChapters := true
  else hasChapters := false;
end;

function TGutenbergTreeView.GetChapter(CharNo : integer) : TTreeNode;
var Middle : integer;
    Low, High : integer;
    CharMiddle : integer;
begin
  if Haschapters then begin
    high      := Items.Count-1;
    low       := 0;
    repeat
      middle := ((high-low) div 2) +Low;
      CharMiddle := PChapter(Items[middle].Data)^.TextIndex;
      if CharNo > CharMiddle then
        Low := middle
      else
        High := middle;
    until Low >= (High - 1);
    result := Items[Low];
  end
  else
    result := nil;
end;

function TGutenbergTreeView.GetIndex;
begin
  Result := PChapter(Selected.data)^.TextIndex;
end;

procedure TGutenbergTreeView.DeleteBookMark(Node : TTreeNode);
begin
  Items.Delete(Node);
end;

function TGutenbergTreeView.AddBookMark(Node : TTreeNode;
         CharNo : integer; Bookmark : String) : TTreeNode;
var PointChapter : PChapter;
    Tn : TTreeNode;
begin
  New(PointChapter);
  PointChapter^.TextIndex := CharNo;
  PointChapter^.Number := -1;
  Tn := Items.AddChild(Node, Bookmark);
  Tn.data := PointChapter;
  result := Tn;
end;



procedure TGutenbergTreeView.SaveTreeToStream(Stream: TStream);
const
  TabChar = #9;
  EndOfLine = #13#10;
var
  i: Integer;
  ANode: TTreeNode;
  NodeStr, DataStr: string;
  Textindex    : String;
begin
  if Items.Count > 0 then
  begin
    ANode := Items[0];
    while ANode <> nil do
    begin
      TextIndex := IntToStr(PChapter(ANode.data)^.TextIndex);
      NodeStr := '';
      DataStr := '';
      for i := 0 to ANode.Level - 1 do begin
          NodeStr := NodeStr + TabChar;

      end;
      NodeStr := NodeStr + ANode.Text + EndOfLine;
      DataStr := DataStr + TextIndex + EndOfLine;
      Stream.Write(Pointer(NodeStr)^, Length(NodeStr));
      Stream.Write(Pointer(DataStr)^, Length(DataStr));
      ANode := ANode.GetNext;
    end;
  end;

  Invalidate;
end;

function TGutenbergTreeView.GetBufStart(Buffer: PChar; var Level: Integer): PChar;
begin
  Level := 0;
  while Buffer^ in [' ', #9] do
  begin
    Inc(Buffer);
    Inc(Level);
  end;
  Result := Buffer;
end;




procedure TGutenbergTreeView.LoadTreeFromStream(Stream: TStream);
var
  List: TStringList;
  ANode, NextNode: TTreeNode;
  ALevel, i: Integer;
  CurrStr, DataStr: string;
  PointChapter : PChapter;
begin
  hasChapters := false;
  List := TStringList.Create;
  Items.BeginUpdate;
  try
    try
      Items.Clear;
      List.LoadFromStream(Stream);
      ANode := nil;
      i := 0;
      while i <= List.Count - 1  do
      begin
        CurrStr := GetBufStart(PChar(List[i]), ALevel);
        if Pos('CHAPTER', CurrStr) > 0 then hasChapters := true;
        DataStr := List[i+1];
        if ANode = nil then
          ANode := Items.AddChild(nil, CurrStr)
        else if ANode.Level = ALevel then
          ANode := Items.AddChild(ANode.Parent, CurrStr)
        else if ANode.Level = (ALevel - 1) then
          ANode := Items.AddChild(ANode, CurrStr)
        else if ANode.Level > ALevel then
        begin
          NextNode := ANode.Parent;
          while NextNode.Level > ALevel do
            NextNode := NextNode.Parent;
          ANode := Items.AddChild(NextNode.Parent, CurrStr);
          New(PointChapter);
          PointChapter^.TextIndex := StrToInt(DataStr);
          ANode.Data := PointChapter;
        end
        else raise ETreeViewError.Create('TreeView Error');
        New(PointChapter);
        PointChapter^.TextIndex := StrToInt(DataStr);
        ANode.Data := PointChapter;
        i := i+2;
      end;
    finally
      Items.EndUpdate;

      List.Free;
      Invalidate;
    end;
  except

    raise;
  end;
end;


procedure TGutenbergTreeView.LoadFromFile(const FileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TGutenbergTreeView.LoadFromStream(Stream: TStream);
begin
  LoadTreeFromStream(Stream);
end;

procedure TGutenbergTreeView.SaveToFile(const FileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TGutenbergTreeView.SaveToStream(Stream: TStream);
begin
  SaveTreeToStream(Stream);
end;

procedure Register;
begin
  RegisterComponents('Gutenberg', [TGutenbergTreeView]);
end;

end.
